from .descriptor import (
    accessor,
    condition_property,
)
from .helper import (
    make_async,
    Registry,
    RegistryMeta,
    async_exit_on_exc,
)
from .wrapper import CoroutineWrapper
from .nursery import (
    TaskNursery,
)

__all__ = [
    "accessor",
    "condition_property",
    "make_async",
    "CoroutineWrapper",
    "TaskNursery",
    "RegistryMeta",
    "Registry",
    "async_exit_on_exc"
]



from __future__ import annotations

import asyncio
import typing

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol

import functools

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property

from .type_vars import T_co, T, T_contra


# pycharm is to stupid to understand this as of now
# https://youtrack.jetbrains.com/issue/PY-29257
class fget_type(Protocol[T_contra]):
    def __call__(self) -> T_contra:
        """
        :class:`fget_type` callables need to have this signature
        """


class fset_type(Protocol[T_co]):
    def __call__(self, value: T_co) -> None:
        """
        :class:`fset_type` callables need to have this signature
        """


class accessor(typing.Generic[T]):
    """
    An accessor provides easy shared access to a resource and

    Example:

        In the simplest case, an accessor synchronizes reads and writes out of the box ::

            >>> import asyncio
            >>> from codestare.async_utils import accessor
            >>> foo = accessor()
            >>> async def wait_for_write(accessor_):
            ...     print(await accessor_.get())
            ...
            >>> background = asyncio.create_task(wait_for_write(foo))
            >>> await foo.set("Bar")
            Bar

        It's possible to use custom getters / setter e.g. create an :class:`accessor` to the value managed
        by a normal property if one needs shared access as well ::

            >>> class Thing:
            ...     def __init__(self):
            ...         self._value = None
            ...     @property
            ...     def value(self):
            ...         return self._value
            ...     @value.setter
            ...     def value(self, val):
            ...         if not val:
            ...             raise ValueError(f"Illegal value {val}")
            ...         self._value = val
            ...
            >>> thing = Thing()
            >>> thing.value = 3
            >>> thing.value
            3
            >>> thing.value = 0
            ValueError: Illegal value 0
            >>> class BetterThing(Thing):
            ...     def __init__(self):
            ...         super().__init__()
            ...         self.value_accessor = accessor(funcs=(
            ...             type(self).value.fget.__get__(self),
            ...             type(self).value.fset.__get__(self)
            ...         ))
            ...
            >>> better_thing = BetterThing()
            >>> background = asyncio.create_task(wait_for_write(better_thing.value_accessor))
            >>> await better_thing.value_accessor.set(3)
            3
            >>> better_thing.value
            3
            >>> await better_thing.value_accessor.set(0)
            ValueError: Illegal value 0

    See Also:
        :class:`condition_property` -- decorator to create `accessor properties` more easily

    """
    fget: 'fget_type[T]'
    fset: 'fset_type[T]'

    @typing.overload
    def __init__(self: accessor[typing.Any],
                 *,
                 funcs: None = ...,
                 condition: asyncio.Condition | None = None,
                 ):
        ...

    @typing.overload
    def __init__(self: accessor[T],
                 *,
                 condition: asyncio.Condition | None = None,
                 ):
        ...

    @typing.overload
    def __init__(self: accessor[T],
                 *,
                 funcs: typing.Tuple[fget_type[T], fset_type[T]] = ...,
                 condition: asyncio.Condition | None = None,
                 ):
        ...

    def __init__(self,
                 *,
                 funcs: typing.Tuple | None = None,
                 condition: asyncio.Condition | None = None,
                 ):
        """
        Args:
            funcs: getter and setter for some value -- `optional`, if not passed get / set a private field of the
                object
            condition: condition to synchronize access to the value -- `optional`, if not passed a new condition
                is created
        """
        if funcs is None:
            def set(instance, value):
                instance._value = value

            def get(instance):
                return instance._value

            self._value = None
            fget = get.__get__(self, type(self))
            fset = set.__get__(self, type(self))
        else:
            non_callable = [f for f in funcs if not callable(f)]
            if non_callable:
                raise ValueError(f"parameters {non_callable} passed as ``funcs`` tuple are not callable")
            fget, fset = funcs

        self.fset = fset
        """
        Setter, either passed via ``funcs`` argument, or a setter of an internal value if no ``funcs`` where passed
        """
        self.fget = fget
        """
        Getter, either passed via ``funcs`` argument, or a getter of an internal value if no ``funcs`` where passed
        """
        self.condition = condition or asyncio.Condition()
        """
        Used to synchronized access, either passed via ``condition`` argument, or a new condition created specifically
        for this accessor
        """

    @property
    def value(self) -> T | None:
        """
        Simple access to the value produced by :attr:`.fget` without async locks i.e. not safe if you did not
        acquire the lock of :attr:`.condition`
        """
        return self.fget()

    async def set(self, value: T) -> None:
        """
        Sets the value (using :attr:`.fset`) and notifies every coroutine waiting on the :attr:`.condition` (e.g.
        :meth:`.get`

        Args:
            value: new value passed to :attr:`.fset`
        """
        async with self.condition:
            self.fset(value)
            self.condition.notify_all()

    async def get(self, *, predicate: typing.Callable[[T], bool] | None = None) -> T:
        """
        Shared access to value produced by :attr:`.fget`

        Args:
            predicate: waits for the predicate result to be truthy, then returns the result of :attr:`.fget`.
                The default predicate (used when ``predicate=None``) returns ``[False, True, True, ...]``, so
                :meth:`.get` blocks once, until it is notified from a :meth:`.set` and then does not block again.
                Passing ``predicate=(lambda: True)`` will make :meth:`.get` not block at all.

        Returns:
            value produced by :attr:`.fget`

        Raises:
            ValueError: if ``predicate`` is not a callable

        See Also:
            :meth:`asyncio.Condition.wait_for` -- used to wait for internal condition
        """
        if predicate is None:
            # this predicate returns False, True i.e. it will block once and always return after notify
            _get_value = iter([False, True]).__next__
            predicate = (
                lambda _: _get_value()
            )

        if not callable(predicate):
            raise ValueError(f"{predicate} is not callable")

        async with self.condition:
            await self.condition.wait_for(lambda: predicate(self.fget()))
            return self.fget()


class condition_property(cached_property, typing.Generic[T]):
    """
    This is a decorator to create a cached :class:`accessor` to handle access to some data via a
    :class:`asyncio.Condition`.

    You can use it like the normal `@property` decorator, but the result of the lookup (`__get__` of the descriptor)
    will be an :class:`accessor` with coroutine attributes to handle safely setting and getting the value
    (from the objects methods passed via ``setter`` and ``getter``, like in normal properties) by means of a condition.

    See Also:
        :class:`accessor` -- how to access the value
    """

    def __init__(self: condition_property[T],
                 fget: typing.Callable[[typing.Any], T] | None = None,
                 fset: typing.Callable[[typing.Any, T], None] | None = None,
                 fdel: typing.Callable[[typing.Any], None] | None = None,
                 doc: str | None = None) -> None:
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

        super().__init__(self._create_accessor)

    def _create_accessor(self: 'condition_property[T]', obj: object) -> accessor[T]:
        return accessor(
            funcs=(functools.partial(self._get, obj), functools.partial(self._set, obj),)
        )

    def _set(self, obj: object, value: T):
        if self.fset is None:
            raise AttributeError(f"can't set attribute {self.attrname}")
        self.fset(obj, value)

    def _get(self, obj: object):
        if self.fget is None:
            raise AttributeError(f'unreadable attribute {self.attrname}')

        if self.fset is None:
            raise AttributeError(f"`get` will block until the next value is set, but no setter is defined.")

        return self.fget(obj)

    def __set__(self, obj, value: T):
        raise AttributeError(f"can't set {self.attrname} directly, use set()")

    @typing.overload
    def __get__(self, instance: None, owner: typing.Type[typing.Any] | None = None) -> condition_property[T]:
        ...

    @typing.overload
    def __get__(self, instance: object, owner: typing.Type[typing.Any] | None = None) -> accessor[T]:
        ...

    def __get__(self, instance: object | None,
                owner: typing.Type[typing.Any] | None = None) -> accessor[T] | condition_property[T]:
        if instance is None:
            return typing.cast(condition_property[T], super().__get__(instance, owner))
        else:
            return typing.cast(accessor[T], super().__get__(instance, owner))

    def getter(self: condition_property[T], fget: typing.Callable[[typing.Any], T]) -> condition_property[T]:
        """
        This is a cached property but uses the same interface as a normal :obj:`property`.
        See example of :obj:`property` documentation on how to use.
        """
        prop = type(self)(fget, self.fset, self.fdel, self.__doc__)
        prop.attrname = self.attrname
        return prop

    def setter(self: condition_property[T], fset: typing.Callable[[typing.Any, T], None]) -> condition_property[T]:
        """
        This is a cached property but uses the same interface as a normal :obj:`property`.
        See example of :obj:`property` documentation on how to use.
        """
        prop = type(self)(self.fget, fset, self.fdel, self.__doc__)
        prop.attrname = self.attrname
        return prop

    def deleter(self: condition_property[T], fdel) -> condition_property[T]:
        """
        This is a cached property but uses the same interface as a normal :obj:`property`.
        See example of :obj:`property` documentation on how to use.
        """
        prop = type(self)(self.fget, self.fset, fdel, self.__doc__)
        prop.attrname = self.attrname
        return prop



from __future__ import annotations

from typing import TypeVar, Coroutine, Any

T = TypeVar('T')
"""
Simple TypeVar
"""
S = TypeVar('S')
"""
Simple TypeVar
"""
R = TypeVar('R')
"""
Simple TypeVar
"""
T_Exception = TypeVar('T_Exception', bound=Exception)
"""
bound to :class:`Exception`
"""
T_co = TypeVar('T_co', covariant=True)
"""
covariant values
"""
T_contra = TypeVar('T_contra', contravariant=True)
"""
contravariant values
"""
SimpleCoroutine = Coroutine[Any, Any, T]
"""
TypeAlias for a Coroutine that only cares about return Type
"""



from __future__ import annotations

import typing

from .type_vars import T, S, R


class CoroutineWrapper(typing.Coroutine[T, S, R]):
    """
    Complex Coroutines are easy to implement with native ``def async`` coroutine syntax, but often require
    some smaller coroutines to compose. Inheriting from :class:`CoroutineWrapper`, a complex coroutine can encapsulate
    all it's dependencies and auxiliary methods.

    Note:
        The example uses an asyncio REPL (``python -m asyncio``)

    Example:

        Let's consider an async generator that generates iterables ::

            >>> async def generate_ranges(up_to: int):
            ...     for n in range(up_to):
            ...             yield range(n)
            ...
            >>> [list(r) async for r in generate_ranges(5)]
            [[], [0], [0, 1], [0, 1, 2], [0, 1, 2, 3]]

        Our coroutine should apply some processing to the values, and to make it interesting it should only
        do so to values matching an additional filter function, i.e. values for which the specified filter function
        returns ``False`` should be ignored ::

            >>> async def coro(process, filter_, ranges):
            ...     async for r in ranges:
            ...         for n in r:
            ...             if not filter_(n): continue
            ...             process(n)
            ...
            >>> await coro(print, lambda n: n % 2 == 0, generate_ranges(5))
            0
            0
            0
            2
            0
            2

        This is pretty concise and one would probably understand what's happening, but for the sake of the
        example we want to break this processing behaviour down into pieces, so that we know
        what's happening. One could e.g. define a custom awaitable ::

            >>> class coro:
            ...     def __init__(self, process, filter_, ranges):
            ...         self.filter = filter_
            ...         self.process = process
            ...         self.ranges = ranges
            ...         self._work = self.work()
            ...     async def make_flat_values(self):
            ...         async for r in self.ranges:
            ...             for n in r:
            ...                 yield n
            ...     async def filter_values(self):
            ...         async for value in self.make_flat_values():
            ...             if self.filter(value):
            ...                 yield value
            ...     async def work(self):
            ...         async for value in self.filter_values():
            ...             self.process(value)
            ...     def __await__(self):
            ...         return self._work.__await__()
            ...
            >>> await coro(print, lambda n: n % 2 == 0, generate_ranges(5))
            0
            0
            0
            2
            0
            2

        The problem is, that ``coro`` now isn't a coroutine, i.e. one can't create a task with it for example ::

            >>> import asyncio
            >>> await asyncio.create_task(coro(print, lambda n: n % 2 == 0, generate_ranges(5)))
            TypeError: a coroutine was expected, got <coro object>

        This is where the :class:`CoroutineWrapper` comes in:

            >>> from codestare.async_utils import CoroutineWrapper
            >>> class coro(CoroutineWrapper):
            ...     def __init__(self, process, filter_, ranges):
            ...         self.filter = filter_
            ...         self.process = process
            ...         self.ranges = ranges
            ...         super().__init__(coroutine=self.work())
            ...     async def make_flat_values(self):
            ...         async for r in self.ranges:
            ...             for n in r:
            ...                 yield n
            ...     async def filter_values(self):
            ...         async for value in self.make_flat_values():
            ...             if self.filter(value):
            ...                 yield value
            ...     async def work(self):
            ...         async for value in self.filter_values():
            ...             self.process(value)
            >>> await asyncio.create_task(coro(print, lambda n: n % 2 == 0, generate_ranges(5)))
            0
            0
            0
            2
            0
            2

        The ``coro`` objects now actually implement the `asyncio` coroutine interface

    """

    def __init__(self: CoroutineWrapper[T, S, R], *, coroutine: typing.Coroutine[T, S, R], **kwargs):
        self._coroutine = coroutine
        super().__init__(**kwargs)

    def __await__(self):
        """
        Proxy to internal coroutine object
        """
        return self._coroutine.__await__()

    def send(self, value):
        """
        Proxy to internal coroutine object
        """
        return self._coroutine.send(value)

    def throw(self, typ, val=None, tb=None):
        """
        Proxy to internal coroutine object
        """
        if val is None:
            return self._coroutine.throw(typ)
        elif tb is None:
            return self._coroutine.throw(typ, val)
        else:
            return self._coroutine.throw(typ, val, tb)

    def close(self):
        """
        Proxy to internal coroutine object
        """
        return self._coroutine.close()



from __future__ import annotations
import typing as t
import asyncio
from contextlib import suppress, AsyncExitStack, asynccontextmanager

import pytest
from codestare.async_utils import TaskNursery, CoroutineWrapper

pytestmark = pytest.mark.asyncio


class MockWrapper(CoroutineWrapper):
    def __init__(self, range: int | None = 20):
        self.value = None
        self.range = range
        super().__init__(coroutine=self._run())

    async def _run(self):
        if self.range is not None:
            for i in range(self.range):
                self.value = i

        else:
            while True:
                self.value += 1


@pytest.fixture
async def nursery(request):
    callback_ran = False

    async def callback():
        nonlocal callback_ran
        callback_ran = True

    nursery = request.param() if hasattr(request, 'param') else TaskNursery()
    nursery.add_sentinel_callback(callback())
    yield nursery

    if any(t.cancelled() for t in nursery._tasks):
        assert callback_ran


@pytest.mark.parametrize('future', [MockWrapper()])
async def test_nursery(future, nursery):
    callback_called = False

    def callback(task):
        nonlocal callback_called
        callback_called = True
        print(f"Callback called for {task}")

    task = nursery.create_task(future)
    task.add_done_callback(callback)
    task.add_done_callback(lambda _: nursery.trigger_sentinel.set())

    task.cancel()
    with suppress(asyncio.CancelledError):
        await task

    assert callback_called


class MockNursery(TaskNursery):
    def __init__(self):
        super().__init__()
        self.exit_stack = AsyncExitStack()
        self._task = None
        self.add_sentinel_callback(self.exit_stack.aclose())
        self._context: t.AsyncContextManager = self.__make_context()

    def run(self):
        if not self._task:
            self._task = self.create_task(MockWrapper(range=None))
            self._task.add_done_callback(lambda _: self.trigger_sentinel.set())

        return self._task

    @asynccontextmanager
    async def __make_context(self):
        task = self.run()
        yield
        task.cancel()
        with suppress(asyncio.CancelledError):
            await task

    def __aenter__(self):
        return self._context.__aenter__()

    def __aexit__(self, exc_type, exc_val, exc_tb):
        return self._context.__aexit__(exc_type, exc_val, exc_tb)


@pytest.mark.parametrize('nursery', [MockNursery], indirect=True)
async def test_exit_stack(nursery: MockNursery):
    value = None

    @asynccontextmanager
    async def test_context_manager():
        nonlocal value
        value = True
        yield
        value = False

    async with nursery:
        await nursery.exit_stack.enter_async_context(test_context_manager())

    assert nursery.run().cancelled()
    assert nursery.trigger_sentinel.is_set()
    await asyncio.sleep(0)
    assert value is not None, "context was not entered"
    assert value is False, "context was not exited"



from __future__ import annotations

from collections import ChainMap

import abc
import asyncio
import sys
import typing
import weakref
from functools import wraps

from .type_vars import T, SimpleCoroutine, S


def make_async(func: typing.Callable[[T], SimpleCoroutine[S] | S]) -> typing.Callable[
    [T], SimpleCoroutine[S]]:
    """
    Decorator to turn a non async function into a coroutine by running it in the default executor pool.
    """

    if asyncio.iscoroutinefunction(func):
        return typing.cast(typing.Callable[[T], SimpleCoroutine[S]], func)

    @wraps(func)
    async def _callback(*args):
        loop = asyncio.get_running_loop()
        result = await loop.run_in_executor(None, func, *args)

        if isinstance(result, typing.Awaitable):
            result = await result

        return result

    return _callback


class RegistryMeta(abc.ABCMeta):
    """
    Set __unique_key_attr__ in created classes to some attribute of the class instance that is
    unique to use this attribute as a key in the registry instead of the __default_key__.
    If the __default_key__ attribute is not present in the instance, it will be copied from the class
    on instance creation, and appended with the number of instances created before.

    All types created by this metaclass will have a ``__registry_key__`` property, so that
    the key for an instance (i.e. either default key or unique key) can be accessed easily.

    Warning:

        If the instances get garbage collected, they will not be available from the registry anymore

    Example:

        Here you can see that the registry only uses weak references, and instances that get garbage collected
        are removed from the registry ::

            >>> from codestare.async_utils import RegistryMeta
            >>> class T(metaclass=RegistryMeta):
            ...     __unique_key_attr__ = 'name'
            ...     def __init__(self, name):
            ...             self.name = name
            ...     def __repr__(self):
            ...             return f"{self.__class__.__name__}(name={self.name!r})"
            ...
            >>> a = T('foo')
            >>> T.registry
            {'foo': T(name='foo')}
            >>> import gc
            >>> gc.collect()
            0
            >>> T.registry
            {'foo': T(name='foo')}
            >>> a = T('bar')
            >>> T.registry
            {'foo': T(name='foo'), 'bar': T(name='bar')}
            >>> gc.collect()
            0
            >>> T.registry
            {'bar': T(name='bar')}

    """
    _wrap_marker = object()

    @staticmethod
    def _post_new(instance: typing.Any):
        cls: RegistryMeta = type(instance)

        # we set value in all parents, so they know a new instance is created
        # we know that the mappings are mutable, so it's ok to set the values.
        for mapping in cls.__created__.maps:
            mapping[instance.__registry_key__] = instance  # type: ignore

    @property
    def registry(cls: typing.Type[T]) -> typing.Dict[typing.Any, T]:
        """
        Mapping :math:`\\text{instance.__registry_key__} \\rightarrow instance`
        """
        return {
            instance.__registry_key__: instance
            for instance in cls.__created__.maps[0].values()
            if instance is not None  # only live references
        }

    @staticmethod
    def _wrap_new(__new__):
        @wraps(__new__)
        def wrapped(cls, *args, **kwargs):
            instance = __new__(cls)
            RegistryMeta._post_new(instance)
            return instance

        wrap_markers = getattr(wrapped, 'markers', set())
        wrap_markers.add(RegistryMeta._wrap_marker)
        wrapped.markers = wrap_markers

        return wrapped

    def __new__(mcs, name, bases, attrs):
        kls: RegistryMeta = super().__new__(mcs, name, bases, attrs)

        wrap_markers = getattr(kls.__new__, 'markers', set())
        if mcs._wrap_marker not in wrap_markers:
            kls.__new__ = mcs._wrap_new(kls.__new__)

        class_namespace = kls.__dict__

        if not hasattr(kls, '__created__'):
            kls.__created__ = ChainMap(weakref.WeakValueDictionary())
        else:
            parent = kls.__created__
            setattr(kls, '__created__', parent.new_child(weakref.WeakValueDictionary()))

        if '__default_key_value__' not in class_namespace:
            kls.__default_key_value__ = f"{attrs['__module__']}.{attrs['__qualname__']}"

        if not hasattr(kls, '__unique_key_attr__'):
            kls.__unique_key_attr__ = '__missing_key__'

        if '__registry_key__' not in class_namespace:
            kls.__registry_key__ = property(
                fget=(
                    lambda instance: getattr(
                        instance,
                        instance.__unique_key_attr__,
                        f"{kls.__default_key_value__}_{id(instance)}"
                    )
                )
            )

        return kls


class Registry(object, metaclass=RegistryMeta):
    """
    You can inherit from this class to implicitly use the :class:`RegistryMeta` metaclass

    Example:

        Define a registry with the metaclass or by inheriting :class:`Registry` ::

            from codestare.async_utils import RegistryMeta, Registry

            # virtually equivalent for most intents and purposes

            class Foo(metaclass=RegistryMeta):
                pass

            class Bar(Registry):
                pass

    See Also:
        :class:`RegistryMeta` -- more information about working with registry classes
    """
    pass


def async_exit_on_exc(ctx_manager: typing.AsyncContextManager, task: asyncio.Task,
                      loop: asyncio.BaseEventLoop = None) -> None:
    """
    Schedules exit of the ``ctx_manager`` if the getting the task result raises an exception other than a
    :class:`asyncio.CancelledError`

    Args:
        ctx_manager: Some context manager that needs to be closed with exception info for exceptions raised
            by the ``task``

        task: a task that maybe succeeded or raised an exception
        loop: event loop to schedule the exit, uses current running loop if not provided -- `optional`

    """

    loop = loop or asyncio.get_running_loop()
    try:
        task.result()
    except asyncio.CancelledError:
        pass
    except:  # noqa
        exc_info = sys.exc_info()
        loop.call_soon(ctx_manager.__aexit__(*exc_info).__await__().__next__)



from __future__ import annotations

import asyncio
import contextlib
import logging
import signal
import sys
import traceback
import typing
import warnings

from . import helper
from .type_vars import T

log = logging.getLogger(__name__)

_TaskYieldType = typing.Optional[asyncio.Future]

__sentinel_tasks__ = set()
WINDOWS = sys.platform == 'win32'


class ExcInfo(typing.NamedTuple):
    """
    As returned by :func:`sys.exc_info`
    """
    exc_type: typing.Optional[typing.Type[Exception]] = None
    exc: typing.Optional[Exception] = None
    traceback: typing.Optional[typing.Any] = None


def mark_sentinel_task(task: asyncio.Task) -> asyncio.Task:
    """
    The package keeps a global reference to all tasks marked as a `sentinel task` i.e. they
    will not be garbage collected even if they are finished. Those tasks will be stopped before
    other tasks if you set up the shutdown handling with this module. These tasks are
    used by the :class:`TaskNursery` to trigger the nursery's shutdown by simply
    canceling / stopping the :attr:`~TaskNursery.sentinel_task`.

    When a :func:`shutdown` task is created (e.g. when a specific signal is received), all sentinel tasks
    will be stopped first -- this will trigger the shutdown of all :class:`TaskNurserys <TaskNursery>` that were
    created in the running interpreter.

    Args:
        task: will be globally referenced to be shut down eventually

    Returns:
        the task
    """
    global __sentinel_tasks__
    __sentinel_tasks__.add(task)
    return task


def handle_exception(loop: asyncio.AbstractEventLoop, context) -> None:
    """
    Calls the loops default exception handler, then creates a :func:`shutdown` task for the loop.
    This is the default exception handler that is set up for a loop with :func:`setup_shutdown_handling` if
    no other specific exception handler is set.

    Args:
        loop: event loop to shut down
        context: meta information for :func:`shutdown` task

    """
    # context["message"] will always be there; but context["exception"] may not
    loop.default_exception_handler(context)

    msg = context.get("tb", context["message"])
    log.error(msg)
    log.info("Shutting down...")

    additional_args = {'name': f"shutdown({context['message']})"} if sys.version_info >= (3, 8) else {}
    asyncio.create_task(shutdown(loop, context=context), **additional_args)


async def stop_task(task: asyncio.Task) -> typing.Any:
    """
    Try to cancel the task, and await it -- returns possible exceptions raised inside the task, e.g. the
    :class:`asyncio.CancelledError` raised by canceling the task prematurely.

    Args:
        task: task to stop, can't be the current running task

    Returns:
        result of task

    Raises:
        ValueError: if the currently running task is passed
    """
    if task is asyncio.current_task():
        raise ValueError(f"Not allowed to stop task running `stop_task`!")

    task.cancel()
    return (await asyncio.gather(task, return_exceptions=True))[0]


def setup_shutdown_handling(loop: asyncio.AbstractEventLoop) -> None:
    """
    Add signal handling and exception handling for the loop.
    On Windows only :obj:`signal.SIGBREAK` and :obj:`signal.SIGINT` are usable, on Linux add handling for
    :obj:`signal.SIGHUP`, :obj:`signal.SIGTERM` and :obj:`signal.SIGINT`.

    Uses normal signal handlers using :mod:`signal` except of :func:`asyncio.loop.add_signal_handler` since the
    latter is not available on Windows systems.

    The signal handlers registered start a :func:`shutdown` task for the passed ``loop``.
    Also sets :func:`handle_exception` as the loop's exception handler if no other handler is already set.

    Args:
        loop: event loop

    """
    _signals = (signal.SIGBREAK, signal.SIGINT) if WINDOWS else (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)

    def _clear_signal_handlers():
        remove_handler = (
            lambda sig: signal.signal(sig, signal.SIG_IGN)
            if WINDOWS else
            loop.remove_signal_handler
        )
        [remove_handler(s) for s in _signals]

    def signal_handler(sig, frame=None):
        kwargs = {} if sys.version_info < (3, 8) else {'name': f'shutdown({sig!r})'}
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError as e:
            warnings.warn(f"Can't create shutdown task - {e}")
        else:
            loop.create_task(shutdown(loop, signal=sig, frame=frame), **kwargs)

    _clear_signal_handlers()

    [signal.signal(sig, signal_handler) for sig in _signals]

    if not loop.get_exception_handler():
        loop.set_exception_handler(handle_exception)


async def shutdown(loop: asyncio.AbstractEventLoop, **kwargs) -> None:
    """
    Await shutdown of the loop.

    If the :func:`shutdown` task is triggered by an exception from a task managed by a :class:`TaskNursery`, the
    ``**kwargs`` will contain a reference to the nursery's :attr:`~TaskNursery.sentinel_task` in the ``context``
    argument. If this is the case, this sentinel task will be stopped (using :func:`stop_task`) which triggers
    trigger the shutdown handling of the :class:`TaskNursery`.

    If the :func:`shutdown` task is triggered by a signal handler, all tasks
    :func:`marked as sentinel tasks <mark_sentinel_task>` will be stopped first, instead.

    Finally all tasks still running in the ``loop`` will be stopped.
    Then the loop will be :meth:`stopped <asyncio.loop.stop>`.

    Args:
        loop: event loop that needs to be stopped
        **kwargs: meta information like received signal or exception context

    """
    sig = kwargs.pop('signal', None)

    if sig:
        log.info(f"Received exit signal {sig!r}...")

    get_tasks = (lambda: [
        task_ for task_ in asyncio.all_tasks()
        if task_ is not asyncio.current_task()
    ])

    context = kwargs.pop('context', {})
    sentinel = context.get('sentinel')
    if sentinel:
        result = await stop_task(sentinel)
    else:
        filter_function = (lambda task_: task_ in __sentinel_tasks__)

        result = await asyncio.gather(
            *map(stop_task, filter(filter_function, get_tasks()))
        )
    if result:
        log.debug(f"Sentinel task[s] stopped with result {result!r}")

    tasks = get_tasks()
    log.info(f"Cancelling {len(tasks)} outstanding tasks")
    for task in tasks:
        await stop_task(task)

    loop.stop()


class TaskNursery(contextlib.AsyncExitStack, helper.Registry):
    """
    Use a :class:`TaskNursery` to create and manage tasks, instead of using the :meth:`asyncio.loop.create_task`
    method directly.

    All tasks created by a nursery receive a callback to execute when they finish, which will trigger
    a graceful shutdown of the event loop if necessary -- this saves a lot of boilerplate code

    Basically, a :class:`TaskNursery` is a :class:`contextlib.AsyncExitStack` that gets closed when it's
    :attr:`~.sentinel_task` is cancelled. This means you can enter any number of async context managers
    with a nursery, and every one of them will be closed when the nursery shuts down.

    Creating a :class:`TaskNursery` for an event loop also sets up signal handling and exception handling
    for that loop, i.e. if a loop has an active :class:`TaskNursery` exiting the interpreter forcefully (e.g.
    ``Ctrl+C`` in the shell) will gracefully shut down all nurseries first.

    :class:`TaskNursery` is a :class:`~codestare.async_utils.helper.Registry`
    with unique attribute :attr:`~TaskNursery.name` by default, i.e. all instances need to have a unique name
    and can be referenced by name using the :attr:`TaskNursery.registry` ::

        >>> from codestare.async_utils import TaskNursery
        >>> import asyncio
        >>> loop = asyncio.get_event_loop_policy().get_event_loop()
        >>> nursery = TaskNursery(name="Foo", loop=loop)
        >>> TaskNursery.registry
        {'Foo': codestare.async_utils.nursery.TaskNursery(name='Foo', loop=<...>)}


    Attributes:

        registry: Inherited from :class:`codestare.async_utils.helper.Registry`

    """
    __has_handling__: typing.Set[asyncio.BaseEventLoop] = set()
    __unique_key_attr__ = 'name'

    @staticmethod
    def add_shutdown_handling(loop):
        if loop not in TaskNursery.__has_handling__:
            setup_shutdown_handling(loop)

    def stop_task(self, task):
        if task not in self._tasks:
            raise ValueError(f"{task} not contained in pending tasks of {self}")

        return stop_task(task)

    def __init__(self, name=None, loop=None):
        super().__init__()
        self._tasks: typing.List[asyncio.Task] = []

        self.fallback_handler = log.exception
        self.loop = loop or asyncio.get_running_loop()

        # teardown behaviour
        self.push_async_callback(self._stop_all)
        self.add_shutdown_handling(self.loop)
        self.sentinel_task = mark_sentinel_task(
            self.create_task(
                self._sentinel_task(event=asyncio.Event()),  # dummy event which is never set
            )
        )
        self.sentinel_task.remove_done_callback(self._task_cb)
        self.name = name or f'TaskNursery-{len(type(self).registry)}'

    @property
    def tasks(self):
        return self._tasks

    async def _sentinel_task(self, event):
        try:
            await event.wait()
        except asyncio.CancelledError:
            pass

        self._tasks.remove(self.sentinel_task)
        self.sentinel_task = None

        await self.aclose()
        return "Success"

    async def _stop_all(self):
        stoppable = [task for task in self._tasks if task is not asyncio.current_task()]
        return await asyncio.gather(*map(self.stop_task, stoppable), return_exceptions=True)

    def _task_cb(self, task: asyncio.Task):
        try:
            result = task.result()
            if task in self._tasks:
                self._tasks.remove(task)
            return result
        except asyncio.CancelledError:
            pass
        except:  # noqa
            exc_info = sys.exc_info()
            self.loop.call_exception_handler(
                {
                    'message': f'{exc_info[1]!r} from {task!r}',
                    'task': task,
                    'tb': traceback.format_exc(),
                    'sentinel': self.sentinel_task
                }
            )

    def create_task(self,
                    coro: typing.Generator[_TaskYieldType, None, T] | typing.Awaitable[T],
                    **kwargs) -> asyncio.Task[T]:

        if sys.version_info >= (3, 8):
            if 'name' not in kwargs:
                kwargs['name'] = getattr(coro, '__name__', str(coro))
            kwargs['name'] += f":{self.__registry_key__}"
        else:
            if kwargs.pop('name', None):
                warnings.warn(f"No `name` argument in {sys.version}")

        task = self.loop.create_task(coro, **kwargs)
        task.add_done_callback(self._task_cb)
        self._tasks.append(task)
        return task

    def __repr__(self):
        kls = self.__class__
        attrs = ['name', 'loop']

        return (f"{kls.__module__}.{kls.__qualname__}("
                f"{', '.join('{}={!r}'.format(name, getattr(self, name, None)) for name in attrs)}"
                f")")
